/*
 * lf3dec.c
 * Decode Leaf LF3 format.
 * 02/05/2000 by TF
 */

#include <stdio.h>
#include <stdlib.h>

#include "lf3.h"

static LF3 *init_lf3(void);
static int read_magic(FILE *);
static void read_size(FILE *, LF3 *);
static void read_unknown(FILE *, LF3 *);
static int extract_body(FILE *, LF3 *);

#define RING_LEN 0x1000

LF3 *read_lf3(char *filename)
{
  LF3 *lf3;
  FILE *fp;

  fp = fopen(filename, "r");
  if (fp == NULL) {
    fprintf(stderr, "read_lf3:Can't open %s.\n", filename);
    return NULL;
  }

  lf3 = init_lf3();
  
  if (read_magic(fp) < 0) {
    fclose(fp);
    return NULL;
  }
  read_size(fp, lf3);
  read_unknown(fp, lf3);

  if (extract_body(fp, lf3) < 0) {
    fclose(fp);
    return NULL;
  }

  fclose(fp);
  return lf3;
}

static LF3 *init_lf3(void)
{
  LF3 *lf3;

  lf3 = (LF3 *)malloc(sizeof(LF3));
  if (lf3 == NULL) {
    fprintf(stderr, "read_lf3: Can't allocate memory.\n");
    return NULL;
  }
  
  return lf3;
}

void free_lf3(LF3 *lf3)
{
  if (lf3 != NULL) {
    if (lf3->image != NULL) {
      free(lf3->image);
    }
    free(lf3);
  }
}


/*
 * Private functions
 */
static int read_magic(FILE *fp)
{
  char buf[9];
  int i;

  for (i = 0; i < 8; i++) {
    buf[i] = fgetc(fp);
  }
  buf[8] = '\0';    /* for safe */
  
  if (strcmp("LEAF64K", buf)) {
    fprintf(stderr, "This file isn't LF3 format.\n");
    return -1;
  }
  return 0;
}

static void read_unknown(FILE *fp, LF3 *lf3)
{
  int i;

  for (i = 0; i < 8; i++) {
    lf3->unknown = fgetc(fp);
    fprintf(stderr, "unknown: %02x(%d)\n", lf3->unknown, lf3->unknown);
  }
}

static void read_size(FILE *fp, LF3 *lf3)
{
  int upper, lower;

  /* x offset */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf3->xoffset = (lower << 8) | upper;

  /* y offset */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf3->yoffset = (lower << 8) | upper;

  /* width */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf3->width = (lower << 8) | upper;

  /* height */
  upper = fgetc(fp);
  lower = fgetc(fp);
  lf3->height = (lower << 8) | upper;

#ifdef DEBUG
  fprintf(stderr, "\noffset %dx%d\n", lf3->xoffset, lf3->yoffset);
  fprintf(stderr, "size %dx%d\n", lf3->width, lf3->height);
#endif

  /* set image size */
  lf3->size = lf3->width * lf3->height * 2;
}

static int extract_body(FILE *fp, LF3 *lf3)
{
  int ring[RING_LEN];
  int i, j;
  int c, m;
  int flag;
  int upper, lower;
  int pos, len;
  int data;
  int idx = 0;
  int r = 0;

  lf3->image = (int *)calloc(lf3->size, sizeof(int));
  if (lf3->image == NULL) {
    fprintf(stderr, "extract_body: Can't allocate memory.\n");
    return -1;
  }
  
  /* initialize ring buffer */
  for (i = 0; i < sizeof(ring) / sizeof(ring[0]); i++){
    /*  ring[i] = -1; */
    ring[i] = 0; /* some images accesses non-initialized memory */
  }
  
  /* extract data */
  for (i = 0, c = 0, m = 0xfee; i < lf3->size;){
    /* flag bits, which indicates data or location */
    if (--c < 0) {
      flag = fgetc(fp);
      r++;
      flag ^= 0xff;
      c = 7;
    }
    
    if(flag & 0x80){
      /* data */
      data = fgetc(fp);
      r++;
      if (data == EOF) {
	fprintf(stderr, "Data may be truncated.\n");
	return -1;
      }
      data ^= 0xff;
      ring[m++] = data;
      lf3->image[idx++] = data;
      i++;
      m &= 0xfff;
    } else {
      /* copy from ring buffer */
      upper = fgetc(fp);
      r++;
      if (upper == EOF) {
	fprintf(stderr, "Data may be truncated.\n");
	return -1;
      }
      upper ^= 0xff;

      lower = fgetc(fp);
      r++;
      if (lower == EOF) {
	fprintf(stderr, "Data may be truncated.\n");
	return -1;
      }
      lower ^= 0xff;
      
      len = (upper & 0x0f) + 3;
      pos = (upper >> 4) + (lower << 4);
#ifdef DEBUG
      if (pos > RING_LEN) {
	fprintf(stderr, "invalid position\n");
	exit(1);
      }
      /* fprintf(stderr, "pos: %x, len: %d\n", pos, len); */
#endif

      for(j = 0; j < len; j++){
	ring[m++] = lf3->image[idx++] =  ring[pos++];
	if (ring[pos - 1] < 0) {
	  fprintf(stderr, "invalid value in ring\n");
	  /* exit(1); */
	}
	i++;
	m &= 0xfff;
	pos &= 0xfff;
      }
    }
    flag = flag << 1;
  }
#ifdef DEBUG
  if (fgetc(fp) != EOF) {
    fprintf(stderr, "Error: Didn't reach EOF.\n");
  } else {
    fprintf(stderr, "Got all data.\n");
  }
#endif

  return 0;
}
